// 
//     MINOTAUR -- It's only 1/2 bull
// 
//     (C)opyright 2009 - 2012 The MINOTAUR Team.
// 

/**
 * \file BndProcessor.h
 * \brief Declare the derived class BndProcessor 
 * \author Ashutosh Mahajan, Argonne National Laboratory
 */

#ifndef MINOTAURBNDPROCESSOR_H
#define MINOTAURBNDPROCESSOR_H

#include "NodeProcessor.h"

namespace Minotaur {

  class Engine;
  class Logger;
  class Solution;
  typedef boost::shared_ptr<Engine> EnginePtr;
  typedef boost::shared_ptr<Logger> LoggerPtr;
  typedef boost::shared_ptr<const Solution> ConstSolutionPtr;

  /**
   * \brief A simple node processor that solves relaxations and branches.
   *
   * BndProcessor solves only continuous relaxations at each node. No
   * heuristic, no presolving, no reformulation. Mainly used for demonstration
   * and debugging.
   */

  class BndProcessor : public NodeProcessor {

    public:
      /// Default constructor
      BndProcessor() { }

      /// Create the processor using options from an environment.
      BndProcessor(EnvPtr env);

      /// Constructor with a given engine.
      BndProcessor(EnginePtr engine, HandlerVector handlers_, LoggerPtr logger);

      /// Destroy.
      ~BndProcessor();
      
      // Base class method.
      Branches getBranches();

      // Base class method.
      WarmStartPtr getWarmStart();

      // Base class method.
      void process(NodePtr node, RelaxationPtr rel, SolutionPoolPtr s_pool);

      // Base class method.
      Bool foundNewSolution(); 

    private:
      /**
       * \brief Check if the current solution is feasible to the original problem.
       * \returns true if the current solution is feasible, false if not.
       * \param[in] node The node currently being processed.
       * \param[in] sol The solution obtained after solving the relaxation at
       *                this node.
       * \param[in] s_pool Solution pool where all solutions are stored.
       * \param[out] should_prune true if the current node can be pruned.
       */
      Bool isFeasible_(NodePtr node, ConstSolutionPtr sol, 
          SolutionPoolPtr s_pool, Bool &should_prune);

      /**
       * \brief Check if current node can be pruned on the basis of solution of the
       * relaxation. 
       *
       * \returns true if this node can be pruned, false otherwise.
       */
      Bool shouldPrune_(NodePtr node, SolutionPoolPtr s_pool);

      /// Solve the relaxation.
      void solveRelaxation_();

      /// Branches found by the brancher for this node.
      Branches branches_;

      /// Engine used to process the relaxation.
      EnginePtr engine_;

      /// Status of the engine.
      EngineStatus engineStatus_;

      /// Pointer to environment.
      EnvPtr env_;

      /// Handlers for checking feasibility etc.
      HandlerVector handlers_;

      /// Log manager.
      LoggerPtr logger_;

      /// How many new solutions were found by the processor.
      UInt numSolutions_;

      /// Relaxation that is processed by this processor.
      RelaxationPtr relaxation_;

      /// Warm-start information for start processing the children.
      WarmStartPtr ws_;
  };

  typedef boost::shared_ptr <BndProcessor> BndProcessorPtr;
}
#endif

// Local Variables: 
// mode: c++ 
// eval: (c-set-style "k&r") 
// eval: (c-set-offset 'innamespace 0) 
// eval: (setq c-basic-offset 2) 
// eval: (setq fill-column 78) 
// eval: (auto-fill-mode 1) 
// eval: (setq column-number-mode 1) 
// eval: (setq indent-tabs-mode nil) 
// End:
