//
//     MINOTAUR -- It's only 1/2 bull
//
//     (C)opyright 2008 - 2012 The MINOTAUR Team.
//

/**
 * \file CxUnivarHandler.h
 * \brief Define the CxUnivarHandler class for handling convex univariate 
 * functions.
 * \author Ashutosh Mahajan, Argonne National Laboratory and Jim Luedtke,
 * UW-Madison
 */

#ifndef MINOTAURCXUNIVARHANDLER_H
#define MINOTAURCXUNIVARHANDLER_H

#include "Handler.h"

namespace Minotaur {

class Engine;
class Function;
class LinearFunction;
class Logger;
class Objective;
class Problem;
typedef boost::shared_ptr<Engine> EnginePtr;
typedef boost::shared_ptr<Function> FunctionPtr;
typedef boost::shared_ptr<LinearFunction> LinearFunctionPtr;
typedef boost::shared_ptr<Logger> LoggerPtr;
typedef boost::shared_ptr<Objective> ObjectivePtr;
typedef boost::shared_ptr<const Problem> ConstProblemPtr;


class CxUnivarConstraintData {

protected:

  /// Input variable, i.e., x in y = f(x)
  /// This should be a pointer to the relaxation variable, look at this to get
  /// local bounds when building or updating a relaxation
  ConstVariablePtr iv_;


  /// Output variable, i.e., y in y = f(x)
  /// This should be a pointer to the relaxation variable
  ConstVariablePtr ov_;

  /// 'L' => y <= f(x), so only add secants
  /// 'G' -> y >= f(x), so only add linearizations
  /// 'E' -> y == f(x), so do both
  char sense_;

  
public:
  
  ConstVariablePtr getOutVar() { return ov_; } ;

  ConstVariablePtr getInputVar() { return iv_; } ;

  char getSense() { return sense_; } ;

   /// Default constructor.
  CxUnivarConstraintData(ConstraintPtr newcon, ConstVariablePtr ovar,
  		ConstVariablePtr ivar, char sense);

  /// Destroy
  ~CxUnivarConstraintData() {};


};

typedef boost::shared_ptr<CxUnivarConstraintData> CxUnivarConstraintDataPtr;
typedef std::vector<CxUnivarConstraintDataPtr> CxUnivarConstraintDataVector;
typedef CxUnivarConstraintDataVector::iterator CxUnivarConstraintIterator;


/**
 * An CxUnivarHandler handles convex univariate functions. The upper relaxation is
 * handled with the single secant inequality, the lower relaxation is handled
 * with linearizations.
 */
class CxUnivarHandler : public Handler {

protected:

  // Creates and adds the secant inequality defined by current constraint to the
  // relaxation
  // Returns the constraint that was added, which may be null if it was not
  // possible to add a constraint (e.g., due to an infinite bound)
  ConstraintPtr addSecant(RelaxationPtr rel, ConstVariablePtr iv,
                          ConstVariablePtr ov, FunctionPtr fn);

  // Creates and adds linearization inequalities to approximate the lower
  // envelope of the convex function
  // Returns a vector of constraints that were added
  ConstraintVector addLin(RelaxationPtr rel, ConstVariablePtr iv,
	    ConstVariablePtr ov, FunctionPtr fn);
 
  /// Tolerance for constraint violation.
  Double eTol_;

  /// Tolerance for when upper and lower bounds considered equal.
  Double vTol_;

  /// Original problem.
  ProblemPtr problem_;

  /// Logger.
  LoggerPtr logger_;
      
  /// For printing.
  static const std::string me_;

  /// Internal data associated with each constraint
  CxUnivarConstraintDataVector cons_data_;

  /// A temporary vector of zeros, for evaluating functions
  DoubleVector tmpX_;
 
  /// A temporary vector of zeros, for getting gradients
  DoubleVector grad;

public:
  /// Default constructor.
  CxUnivarHandler(EnvPtr env, ProblemPtr problem);

  /// Destroy
  ~CxUnivarHandler();

  /**
   * Adds constraint to list (as all handlers), but also constructs the
   * associated constraint data.
   */
  void addConstraint(ConstraintPtr newcon, ConstVariablePtr ivar,
                     ConstVariablePtr ovar, char sense = 'E'); 

  // base class method.
  void addConstraint(ConstraintPtr ) {};

  /**
   *  For this handler, nothing is different at root or any node when doing full
   *  relax
   */
  void relaxInitFull(RelaxationPtr rel, Bool *isInfeasible) {
    relaxNodeFull(NodePtr(), rel, isInfeasible);
  }

  void relaxInitInc(RelaxationPtr /* rel */, Bool* /* isInfeasible */) {} 

  /**
   * Check feasibility.
   */
  Bool isFeasible(ConstSolutionPtr sol, RelaxationPtr relaxation, 
                  Bool & isInfeasible);

  /**
   * Not implemented yet. Eventually, could add violated linearization
   * inequalities for underestimator portion
   */
  void separate(ConstSolutionPtr sol, NodePtr node, RelaxationPtr rel,
                SolutionPoolPtr s_pool, Bool *sol_found,
                SeparationStatus *status);


  /** 
   * Create a relaxation by adding the secant inequality for the upper estimator,
   * and some number of linearization inequalities, at a minimum from the end
   * points
   */
  virtual void relaxNodeFull(NodePtr node, RelaxationPtr rel, Bool *isInfeasible); 

  /** 
   * Create a relaxation by updating the secant inequality for the upper estimator,
   * and adding lineariations at the end points, if they are new 
   */
  virtual void relaxNodeInc(NodePtr /* node */, RelaxationPtr /* rel */,
                            Bool * /* isInfeasible */) {} ;

  /// Not implemented 
  virtual void getBranchingCandidates(RelaxationPtr rel, 
                                      const std::vector< Double > &x, ModVector & mods, BrCandSet & cands, 
                                      Bool & isInf);

  // Implement Handler::getBrMod().
  virtual ModificationPtr getBrMod(BrCandPtr cand, DoubleVector &x, 
                                   RelaxationPtr rel, BranchDirection dir);

  // Implement Handler::getBranches().
  virtual Branches getBranches(BrCandPtr cand, DoubleVector & x,
                               RelaxationPtr rel, SolutionPoolPtr s_pool);

  // presolve.
  virtual SolveStatus presolve(PreModQ *pre_mods);

  // Implement Handler::presolveNode().
  virtual Bool presolveNode(ProblemPtr p, NodePtr node,
                            SolutionPoolPtr s_pool, ModVector &n_mods,
                            ModVector &t_mods);
                            

  // Write name
  virtual std::string getName() const;

};

/// Shared pointer to CxUnivarHandler.
typedef boost::shared_ptr<CxUnivarConstraintData> CxUnivarConstraintDataPtr;

/// Shared pointer to const CxUnivarHandler.
typedef boost::shared_ptr<const CxUnivarConstraintData> CxUnivarConstConstraintDataPtr;

/// Shared pointer to CxUnivarHandler.
typedef boost::shared_ptr<CxUnivarHandler> CxUnivarHandlerPtr;

/// Shared pointer to const CxUnivarHandler.
typedef boost::shared_ptr<const CxUnivarHandler> CxUnivarConstHandlerPtr;
}
#endif

// Local Variables: 
// mode: c++ 
// eval: (c-set-style "k&r") 
// eval: (c-set-offset 'innamespace 0) 
// eval: (setq c-basic-offset 2) 
// eval: (setq fill-column 78) 
// eval: (auto-fill-mode 1) 
// eval: (setq column-number-mode 1) 
// eval: (setq indent-tabs-mode nil) 
// End:
