//
// MINOTAUR -- It's only half bull!
//
// (C)opyright 2009 - 2012 The MINOTAUR Team.
//

/**
 * \file Environment.h
 * \brief Define the Environment class.
 * \author Ashutosh Mahajan, Argonne National Laboratory
 */

#ifndef ENVIRONMENT_H
#define ENVIRONMENT_H

#include "Types.h"

namespace Minotaur {

  class Interrupt;
  class Logger;
  class Timer;
  class TimerFactory;
  typedef boost::shared_ptr<Logger> LoggerPtr;

  /**
   * The environment is a container class that has pointers to the
   * Logger, interrupt handler, timer factory and options that need to be
   * passed to Minotaur.
   */
  class Environment {
    public:
      /// Default constructor.
      Environment();

      /// Destroy.
      ~Environment();

      /// Read the options using a char array that has 'argc' words in it.
      void readOptions(const Int argc, char **argv);

      /// Read options from a string.
      void readOptions(const std::string &str);
      
      /// Get an interrupt.
      Interrupt *getInterrupt() const { return 0;};

      /// Get the logger that is setup for the whole environment.
      LoggerPtr getLogger() const;

      /// Get a new timer. The calling function has to free this timer.
      Timer *getTimer();

      /** 
       * Write the full information about the version number and how Minotaur
       * was compiled. TODO: Fix it.
       */
      void writeFullVersion(std::ostream &out);

      /// Get the version string
      std::string getVersion();

      /// Get the options database.
      OptionDBPtr getOptions();

    private:
      /// The options database
      OptionDBPtr options_;

      /// The generator that is used to build timers.
      TimerFactory *timerFac_;

      /// Interrupt?
      Interrupt *interrupt_;

      /// Logger that is used for the whole environment.
      LoggerPtr logger_;

      /// Create default set of options
      void createDefaultOptions_();

      /// For log:
      static const std::string me_;

      /// Set a new interrupt.
      void setInterrupt(Interrupt *i);

      /// Set a new logger.
      void setLogger(LoggerPtr l);

      /// Remove the string "minotaur." from front of a string.
      void removeMinotaurPrepend_(std::string & fullname);

      /**
       * Remove leading dashes "--" or "-" from front of a string. Returns
       * the number of dashes removed. The string "name" is modified.
       */
      UInt removeDashes_(std::string &name);

      /**
       * Reach options from config file and increment the number of
       * instances.
       */
      void readConfigFile_(std::string fname, UInt &num_p);

      /**
       * Return true if the string "str" corresponds to true. ("T", "TRUE",
       * "Y", "YES", "1", or their lower case equivalents). Returns false
       * otherwise.
       */
      Bool getBoolValue_(const std::string & str);

      /*
       * Find an option with the provided name. The option is option of that
       * type is returned and all others are NULLed.
       */
      void findOption_(const std::string &name, BoolOptionPtr &b_option,
          IntOptionPtr &i_option, DoubleOptionPtr &d_option, 
          StringOptionPtr &s_option, FlagOptionPtr &f_option);

      /// Add an option to the database.
      void convertAndAddOneOption_(BoolOptionPtr &b_option, 
          IntOptionPtr &i_option, DoubleOptionPtr &d_option, 
          StringOptionPtr &s_option, FlagOptionPtr &f_option, std::string &name,
          std::string &value);
        
      /**
       * Remove the substring that follows '=' sign in 'name' and return the
       * substring.
       */
      std::string separateEqualToArg_(std::string &name);
  };
}
#endif

// Local Variables: 
// mode: c++ 
// eval: (c-set-style "k&r") 
// eval: (c-set-offset 'innamespace 0) 
// eval: (setq c-basic-offset 2) 
// eval: (setq fill-column 78) 
// eval: (auto-fill-mode 1) 
// eval: (setq column-number-mode 1) 
// eval: (setq indent-tabs-mode nil) 
// End:
