//
//     MINOTAUR -- It's only 1/2 bull
//
//     (C)opyright 2008 - 2012 The MINOTAUR Team.
//

/**
 * \file Transformer.h
 * \brief Declare base class for transforming problems.
 * \author Ashutosh Mahajan, Argonne National Laboratory
 */

#ifndef MINOTAURTRANSFORMER_H
#define MINOTAURTRANSFORMER_H

#include "Types.h"
#include "OpCode.h"

namespace Minotaur {
class CxQuadHandler;
class CxUnivarHandler;
class CGraph;
class CNode;
class Environment;
class LinearHandler;
class Logger;
class Problem;
class Solution;
typedef boost::shared_ptr<CxQuadHandler> CxQuadHandlerPtr;
typedef boost::shared_ptr<CxUnivarHandler> CxUnivarHandlerPtr;
typedef boost::shared_ptr<CGraph> CGraphPtr;
typedef boost::shared_ptr<Environment> EnvPtr;
typedef boost::shared_ptr<LinearHandler> LinearHandlerPtr;
typedef boost::shared_ptr<Logger> LoggerPtr;
typedef boost::shared_ptr<Problem> ProblemPtr;
typedef boost::shared_ptr<Solution> SolutionPtr;
typedef boost::shared_ptr<const Solution> ConstSolutionPtr;


// Container for y = lf(x).
class YEqExprs
{
public:
  YEqExprs(UInt ) {};
  VariablePtr findY(CGraphPtr ) { return VariablePtr(); };
  void insert(VariablePtr , VariablePtr , VariablePtr ) {};

private:
  UInt n_;
  DoubleVector hash_;
  std::vector<OpCode> op_;
  VarVector x1_;
  VarVector x2_;
  VarVector y_;
};


class YEqLfs
{
public:
  YEqLfs(UInt ) {};
  VariablePtr findY(LinearFunctionPtr, Double) { return VariablePtr(); };
  void insert(VariablePtr , LinearFunctionPtr , Double ) {};

private:
  UInt n_;
  DoubleVector hash_;
  std::vector<LinearFunctionPtr> lf_;
  DoubleVector d_;
  VarVector y_;
};


class YEqVars
{
public:
  YEqVars(UInt ) {};
  VariablePtr findY(VariablePtr , Double ) { return VariablePtr(); };
  void insert(VariablePtr , VariablePtr , Double ) {};

private:
  DoubleVector d_;
  DoubleVector hash_;
  UInt n_;
  std::vector<LinearFunctionPtr> x_;
  VarVector y_;
};


/**
 * \brief Base class for reformulating a problem so that handlers can be
 * applied to it.
 *
 * A transformer will create a new problem equivalent to a given problem by
 * spliting constraints, adding new variables etc. The end result is a
 * problem whose each constraint can be handled by a specific handler.
 */
class Transformer {
public:

  /// Default Constructor.
  Transformer();

  /// Constructor.
  Transformer(EnvPtr env);

  /// Destroy.
  virtual ~Transformer();

  void clearUnusedHandlers_(HandlerVector &handlers);
  void copyLinear(ConstProblemPtr p, ProblemPtr newp);
  void copyVars(ConstProblemPtr p, ProblemPtr newp);


  /// Translate the solution of reformulated problem into that of original
  /// problem.
  virtual SolutionPtr getSolOrig(ConstSolutionPtr sol);

  /// Translate the solution of reformulated problem into that of original
  /// problem.
  virtual SolutionPtr getSolReform(ConstSolutionPtr sol);

  virtual void refNonlinCons(ConstProblemPtr oldp, ProblemPtr newp);
    
  virtual void refNonlinObj(ConstProblemPtr oldp, ProblemPtr newp);
    
  /// Do the reformulation, and assign handlers.  Base implementation is empty.
  virtual void reformulate(ConstProblemPtr oldp, ProblemPtr &newp, 
                           HandlerVector &handlers, Int &status);

  /**
     Takes Problem pointed to in p, and turns it into one with a linear
     objective by adding a new variable if necessary 
  */
  virtual void makeObjectiveLinear(ProblemPtr newp);


protected:
  /// The pointer to environment.
  EnvPtr env_;

  /// Handler for quadratic terms
  CxQuadHandlerPtr qHandler_;

  /// Handler for univariate constraints with convex shape.
  CxUnivarHandlerPtr uHandler_;

  /// Handler for linear constraints and variables.
  LinearHandlerPtr lHandler_;

  LoggerPtr logger_;

  static const std::string me_;
  DoubleVector rand_;
  YEqVars *yVars_;
  YEqLfs *yLfs_;
  YEqExprs *yExprs_;

  const Double zTol_;

  /**
   * \brief Check if all constraints in a problem have been assigned to a handler.
   * \param[in] p Problem whose constraints need to be checked.
   * \returns True if all constraints have been assigned. False otherwise.
   */
  Bool allConsAssigned_(ProblemPtr p, HandlerVector &handlers);
  void assignHandler_(CGraphPtr cg, ConstraintPtr c);

  void chkLfSing_(LinearFunctionPtr lf, VariablePtr *v, Double *d);

  /// Convert a maximization objective into minimization.
  void minObj_();
  VariablePtr newVar_(VariablePtr iv, Double d, ProblemPtr newp);
  VariablePtr newVar_(LinearFunctionPtr lf, Double d, ProblemPtr newp);
  VariablePtr newVar_(CGraphPtr cg, ProblemPtr newp);
  void recursRef_(const CNode *node, CGraphPtr &cgout, LinearFunctionPtr &lf,
                  VariablePtr &v, Double &d, ProblemPtr newp);
  void setup_(UInt n);
  void uniVarRef_(OpCode op, CGraphPtr cgl, LinearFunctionPtr lfl, 
                  VariablePtr vl, Double dl, ProblemPtr newp, CGraphPtr &cg,
                  LinearFunctionPtr &lf, VariablePtr &v, Double &d);
};

typedef boost::shared_ptr<Transformer> TransformerPtr;
typedef boost::shared_ptr<const Transformer> ConstTransformerPtr;

}

#endif


// Local Variables: 
// mode: c++ 
// eval: (c-set-style "k&r") 
// eval: (c-set-offset 'innamespace 0) 
// eval: (setq c-basic-offset 2) 
// eval: (setq fill-column 78) 
// eval: (auto-fill-mode 1) 
// eval: (setq column-number-mode 1) 
// eval: (setq indent-tabs-mode nil) 
// End:
